@@ -244,17 +244,21 @@ int ClientHandler::write_tls() {
 
   ERR_clear_error();
 
+  if (on_write() != 0) {
+    return -1;
+  }
+
+  auto iovcnt = upstream_->response_riovec(&iov, 1);
+  if (iovcnt == 0) {
+    conn_.start_tls_write_idle();
+
+    conn_.wlimit.stopw();
+    ev_timer_stop(conn_.loop, &conn_.wt);
+
+    return 0;
+  }
+
   for (;;) {
-    if (on_write() != 0) {
-      return -1;
-    }
-
-    auto iovcnt = upstream_->response_riovec(&iov, 1);
-    if (iovcnt == 0) {
-      conn_.start_tls_write_idle();
-      break;
-    }
-
     auto nwrite = conn_.write_tls(iov.iov_base, iov.iov_len);
     if (nwrite < 0) {
       return -1;
@@ -265,12 +269,12 @@ int ClientHandler::write_tls() {
     }
 
     upstream_->response_drain(nwrite);
+
+    iovcnt = upstream_->response_riovec(&iov, 1);
+    if (iovcnt == 0) {
+      return 0;
+    }
   }
-
-  conn_.wlimit.stopw();
-  ev_timer_stop(conn_.loop, &conn_.wt);
-
-  return 0;
 }
 
 int ClientHandler::upstream_noop() { return 0; }
@@ -643,7 +647,7 @@ void ClientHandler::pool_downstream_connection(
 
   dconn->set_client_handler(nullptr);
 
-  auto group = dconn->get_downstream_addr_group();
+  auto &group = dconn->get_downstream_addr_group();
 
   if (LOG_ENABLED(INFO)) {
     CLOG(INFO, this) << "Pooling downstream connection DCONN:" << dconn.get()
@@ -702,8 +706,21 @@ Http2Session *ClientHandler::select_http2_session_with_affinity(
                      << ", index=" << (addr - shared_addr->addrs.data());
   }
 
-  if (addr->http2_extra_freelist.size()) {
-    auto session = addr->http2_extra_freelist.head;
+  for (auto session = addr->http2_extra_freelist.head; session;) {
+    auto next = session->dlnext;
+
+    if (session->max_concurrency_reached(0)) {
+      if (LOG_ENABLED(INFO)) {
+        CLOG(INFO, this)
+            << "Maximum streams have been reached for Http2Session(" << session
+            << ").  Skip it";
+      }
+
+      session->remove_from_freelist();
+      session = next;
+
+      continue;
+    }
 
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "Use Http2Session " << session
@@ -766,34 +783,73 @@ Http2Session *ClientHandler::select_http2_session(
   auto &http2_avail_freelist = shared_addr->http2_avail_freelist;
 
   if (http2_avail_freelist.size() >= min) {
-    auto session = http2_avail_freelist.head;
-    session->remove_from_freelist();
+    for (auto session = http2_avail_freelist.head; session;) {
+      auto next = session->dlnext;
 
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "Use Http2Session " << session
-                       << " from http2_avail_freelist";
-    }
+      session->remove_from_freelist();
 
-    if (session->max_concurrency_reached(1)) {
-      if (LOG_ENABLED(INFO)) {
-        CLOG(INFO, this) << "Maximum streams are reached for Http2Session("
-                         << session << ").";
+      // session may be in graceful shutdown period now.
+      if (session->max_concurrency_reached(0)) {
+        if (LOG_ENABLED(INFO)) {
+          CLOG(INFO, this)
+              << "Maximum streams have been reached for Http2Session("
+              << session << ").  Skip it";
+        }
+
+        session = next;
+
+        continue;
       }
-    } else {
-      session->add_to_avail_freelist();
+
+      if (LOG_ENABLED(INFO)) {
+        CLOG(INFO, this) << "Use Http2Session " << session
+                         << " from http2_avail_freelist";
+      }
+
+      if (session->max_concurrency_reached(1)) {
+        if (LOG_ENABLED(INFO)) {
+          CLOG(INFO, this) << "Maximum streams are reached for Http2Session("
+                           << session << ").";
+        }
+      } else {
+        session->add_to_avail_freelist();
+      }
+      return session;
     }
-    return session;
   }
 
   DownstreamAddr *selected_addr = nullptr;
 
   for (auto &addr : shared_addr->addrs) {
-    if (addr.proto != PROTO_HTTP2 || (addr.http2_extra_freelist.size() == 0 &&
-                                      addr.connect_blocker->blocked())) {
+    if (addr.in_avail || addr.proto != PROTO_HTTP2 ||
+        (addr.http2_extra_freelist.size() == 0 &&
+         addr.connect_blocker->blocked())) {
       continue;
     }
 
-    if (addr.in_avail) {
+    for (auto session = addr.http2_extra_freelist.head; session;) {
+      auto next = session->dlnext;
+
+      // session may be in graceful shutdown period now.
+      if (session->max_concurrency_reached(0)) {
+        if (LOG_ENABLED(INFO)) {
+          CLOG(INFO, this)
+              << "Maximum streams have been reached for Http2Session("
+              << session << ").  Skip it";
+        }
+
+        session->remove_from_freelist();
+
+        session = next;
+
+        continue;
+      }
+
+      break;
+    }
+
+    if (addr.http2_extra_freelist.size() == 0 &&
+        addr.connect_blocker->blocked()) {
       continue;
     }
 
@@ -1143,7 +1199,8 @@ void ClientHandler::write_accesslog(Downstream *downstream) {
   upstream_accesslog(
       get_config()->logging.access.format,
       LogSpec{
-          downstream, StringRef{ipaddr_}, http2::to_method_string(req.method),
+          downstream, downstream->get_addr(), StringRef{ipaddr_},
+          http2::to_method_string(req.method),
 
           req.method == HTTP_CONNECT
               ? StringRef(req.authority)
@@ -1176,7 +1233,7 @@ void ClientHandler::write_accesslog(int major, int minor, unsigned int status,
 
   upstream_accesslog(get_config()->logging.access.format,
                      LogSpec{
-                         nullptr, StringRef(ipaddr_),
+                         nullptr, nullptr, StringRef(ipaddr_),
                          StringRef::from_lit("-"), // method
                          StringRef::from_lit("-"), // path,
                          StringRef(alpn_), nghttp2::ssl::get_tls_session_info(
@@ -1444,4 +1501,10 @@ StringRef ClientHandler::get_forwarded_for() const {
 
 const UpstreamAddr *ClientHandler::get_upstream_addr() const { return faddr_; }
 
+Connection *ClientHandler::get_connection() { return &conn_; };
+
+void ClientHandler::set_tls_sni(const StringRef &sni) { sni_ = sni.str(); }
+
+StringRef ClientHandler::get_tls_sni() const { return StringRef{sni_}; }
+
 } // namespace shrpx
